//
// Fl_Table -- A table widget
//
// Copyright 2002 by Greg Ercolano.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Library General Public
// License as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Library General Public License for more details.
//
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA.
//
// Please report all bugs and problems to "erco@seriss.com".
//
// TODO:
//    o Auto scroll during dragged selection
//    o Keyboard navigation (up/down/left/right arrow)
//    o Row headings (only column headings supported currently)
//
// BUGS:
//    o Dragging selection beyond edge of widget causes a 'wild select'
//

#include <FL/fl_draw.H>
#include "Fl_Table.H"

#define SCROLLBAR_SIZE	16

// Insure scrollbars receive events by making them last children
//    (Same technique used by Fl_Scroll in fltk 1.1.3)
//
void Fl_Table::fix_scrollbar_order()
{
    Fl_Widget **a = (Fl_Widget**)array();
    if ( a[children()-1] != vscrollbar )
    {
        int i, j;
	for ( i=j=0; j<children(); j++ )
	{
	    if ( a[j] != hscrollbar && a[j] != vscrollbar )
	        a[i++] = a[j];
	}
	a[i++] = hscrollbar;
	a[i++] = vscrollbar;
    }
}

// Scroll display so 'row' is at top
void Fl_Table::row_position(int row)
{
    if ( row < 0 ) row = 0;
    else if ( row >= rows() ) row = rows() - 1;
    double newtop = row_scroll_position(row);
    if ( newtop > vscrollbar->maximum() ) 
        { newtop = vscrollbar->maximum(); }
    vscrollbar->Fl_Slider::value(newtop);
    table_scrolled();
    redraw();
}

// Scroll display so 'col' is at left
void Fl_Table::col_position(int col)
{
    if ( col < 0 ) col = 0;
    else if ( col >= cols() ) col = cols() - 1;
    double newleft = col_scroll_position(col);
    if ( newleft > hscrollbar->maximum() ) 
        { newleft = hscrollbar->maximum(); }
    hscrollbar->Fl_Slider::value(newleft);
    table_scrolled();
    redraw();
}

// Find scroll position of a row (in pixels)
long Fl_Table::row_scroll_position(int row)
{
    long scroll = 0;
    for ( int t=0; t<row; t++ )
        scroll += row_height(t);
    return(scroll);
}

// Find scroll position of a column (in pixels)
long Fl_Table::col_scroll_position(int col)
{
    long scroll = 0;
    for ( int t=0; t<col; t++ )
        scroll += col_width(t);
    return(scroll);
}

// Ctor
Fl_Table::Fl_Table(int X, int Y, int W, int H, const char *l) : Fl_Group(X,Y,W,H,l)
{
    _rows             = 0;
    _cols             = 0;
    _row_header_w     = 40;
    _col_header_h     = 18;
    _row_header       = 0;
    _col_header       = 0;
    _row_header_color = color();
    _col_header_color = color();
    _row_resize       = 0;
    _col_resize       = 0;
    _tablew           = 0;
    _tableh           = 0;
    toprow            = 0;
    botrow            = 0;
    leftcol           = 0;
    rightcol          = 0;
    last_cursor       = FL_CURSOR_DEFAULT;
    _when             = FL_WHEN_RELEASE;

    box(FL_THIN_DOWN_FRAME);

    tablebox = new Fl_Box(x(), y(), w(), h());
    tablebox->box(FL_NO_BOX);

    vscrollbar = new Fl_Scrollbar(x()+w()-SCROLLBAR_SIZE, y(), SCROLLBAR_SIZE, h()-SCROLLBAR_SIZE);
    vscrollbar->type(FL_VERTICAL);
    vscrollbar->callback(scroll_cb, (void*)this);

    hscrollbar = new Fl_Scrollbar(x(), y()+h()-SCROLLBAR_SIZE, w(), SCROLLBAR_SIZE);
    hscrollbar->type(FL_HORIZONTAL);
    hscrollbar->callback(scroll_cb, (void*)this);

    table_resized();

    end();
}

// Dtor
Fl_Table::~Fl_Table()
{
}

// Set height of a row
void Fl_Table::row_height(int row, int height)
{
    if ( row < 0 || row >= rows() ) return;
    _rowheights[row] = height;
    table_resized();
}

// Set width of a column
void Fl_Table::col_width(int col, int width)
{
    if ( col < 0 || col >= _cols ) return;
    _colwidths[col] = width;
    table_resized();
}

// Return row/col clamped to reality
int Fl_Table::row_col_clamp(int &R, int &C)
{
    int clamped = 0;
    if ( R < 0 )      { R = 0;       clamped = 1; }
    if ( R >= _rows ) { R = _rows-1; clamped = 1; }
    if ( C < 0 )      { C = 0;       clamped = 1; }
    if ( C >= _cols ) { C = _cols-1; clamped = 1; }
    return(clamped);
}

// Return bounding region for given context
void Fl_Table::get_bounds(TableContext context, int &X, int &Y, int &W, int &H)
{
    switch ( context )
    {
        case CONTEXT_COL_HEADER:

	    // Column header clipping.
	    X = tablebox->x();
	    Y = tablebox->y() - col_header_height();
	    W = tablebox->w();
	    H = col_header_height();
	    return;

        case CONTEXT_ROW_HEADER:

	    // Row header clipping.
	    X = tablebox->x() - row_header_width();
	    Y = tablebox->y();
	    W = row_header_width();
	    H = tablebox->h();
	    return;

	case CONTEXT_TABLE:

	    // Table window clipping.
	    //     If table window touches any Fl_Table widget edge, 
	    //     compensate for the 1 pixel border all around its inside.
	    //
	    X = tablebox->x();
	    Y = tablebox->y();
	    W = tablebox->w();
	    H = tablebox->h();
	    return;

	// TODO: Add other contexts..
	default:
	    // fprintf(stderr, "get_bounds(): context %d unimplemented\n", context);
	    return;
    }
    //NOTREACHED
}

// Find row/col beneath cursor
//
//    Returns R/C and context.
//    Also returns resizeflag, if mouse is hovered over a resize boundary.
//
Fl_Table::TableContext Fl_Table::cursor2rowcol(int &R, int &C, ResizeFlag &resizeflag)
{
    // return values
    R = C = 0;
    resizeflag = RESIZE_NONE;

    // Event in scrollbar?
    //    Then event is not in table context.
    //
    if ( ( vscrollbar->visible() && Fl::event_inside(vscrollbar) ) || 
         ( hscrollbar->visible() && Fl::event_inside(hscrollbar) ) )
	{ return(CONTEXT_NONE); }
    
    int X, Y, W, H;

    // Row header?
    if ( row_header() )
    {
	get_bounds(CONTEXT_ROW_HEADER, X, Y, W, H);

	// Inside a row heading?
	if ( Fl::event_inside(X, Y, W, H) )
	{
	    // Scan visible rows until found
	    for ( R = toprow; R <= botrow; R++ )
	    {
		find_cell(CONTEXT_ROW_HEADER, R, 0, X, Y, W, H);
		if ( Fl::event_y() >= Y && Fl::event_y() < (Y+H) )
		{
		    // Found row?
		    //     If cursor over resize boundary, and resize enabled,
		    //     enable the appropriate resize flag.
		    //
		    if ( row_resize() )
		    {
		        if ( Fl::event_y() <= (Y+3-0) ) { resizeflag = RESIZE_ROW_ABOVE; }
		        if ( Fl::event_y() >= (Y+H-3) ) { resizeflag = RESIZE_ROW_BELOW; }
		    }
		    return(CONTEXT_ROW_HEADER);
		}
	    }

	    // Must be in row header dead zone
	    return(CONTEXT_NONE);
	}
    }

    // Column header?
    if ( col_header() )
    {
        get_bounds(CONTEXT_COL_HEADER, X, Y, W, H);

	// Inside a column heading?
	if ( Fl::event_inside(X, Y, W, H) )
	{
	    // Scan visible columns until found
	    for ( C = leftcol; C <= rightcol; C++ )
	    {
		find_cell(CONTEXT_COL_HEADER, 0, C, X, Y, W, H);
		if ( Fl::event_x() >= X && Fl::event_x() < (X+W) )
		{
		    // Found column?
		    //     If cursor over resize boundary, and resize enabled,
		    //     enable the appropriate resize flag.
		    //
		    if ( col_resize() )
		    {
		        if ( Fl::event_x() <= (X+3-0) ) { resizeflag = RESIZE_COL_LEFT; }
		        if ( Fl::event_x() >= (X+W-3) ) { resizeflag = RESIZE_COL_RIGHT; }
		    }
		    return(CONTEXT_COL_HEADER);
		}
	    }

	    // Must be in column header dead zone
	    return(CONTEXT_NONE);
	}
    }

    // Mouse somewhere in table?
    //     Scan visible r/c's until we find it.
    //
    for ( R = toprow; R <= botrow; R++ )
    {
	for ( C = leftcol; C <= rightcol; C++ )
	{
	    find_cell(CONTEXT_CELL, R, C, X, Y, W, H);
	    if ( Fl::event_inside(X, Y, W, H) )
	        { return(CONTEXT_CELL); }	// found it
	}
    }

    // Somewhere else in table
    return(CONTEXT_NONE);
}

// Find X/Y/W/H for cell at R/C
//     If R or C are out of range, returns -1 
//     with X/Y/W/H set to zero.
//
int Fl_Table::find_cell(TableContext context, int R, int C, int &X, int &Y, int &W, int &H)
{
    if ( row_col_clamp(R, C) )
	{ X=Y=W=H=0; return(-1); }

    X = col_scroll_position(C) - hscrollbar->value() + tablebox->x();
    Y = row_scroll_position(R) - vscrollbar->value() + tablebox->y();
    W = col_width(C);
    H = row_height(R);

    switch ( context )
    {
	case CONTEXT_COL_HEADER:
	    Y = tablebox->y() - col_header_height();
	    H = col_header_height();
	    return(0);

	case CONTEXT_ROW_HEADER:
	    X = tablebox->x() - row_header_width();
	    W = row_header_width();
	    return(0);

	case CONTEXT_CELL:
	    return(0);

	case CONTEXT_TABLE:
	    return(0);

	// TODO -- HANDLE OTHER CONTEXTS
	default:
	    // fprintf(stderr, "find_cell: unknown context %d\n", (int)context);
	    return(-1);
    }
    //NOTREACHED
}

// Return table window size
void Fl_Table::table_window_size(int &X, int &Y, int &W, int &H)
{
    X=x(); Y=y(); H=h(); W=w();

    // Trim window if scrollbars visible
    if ( vscrollbar->visible() ) { W -= SCROLLBAR_SIZE; }
    if ( hscrollbar->visible() ) { H -= SCROLLBAR_SIZE; }

    // Trim window if headers enabled
    if ( col_header() )
    {
        Y += col_header_height();
	H -= col_header_height();
    }
    if ( row_header() )
    {
        X += row_header_width();
	W -= row_header_width();
    }
}

// Recalculate internals after a scroll.
//
//    Call this if table has been scrolled or resized.
//    Does not handle redraw().
//    TODO: Assumes tablebox has already been recalculated.
//
void Fl_Table::table_scrolled()
{
    // Nothing to see?
    if ( _rows == 0 || _cols == 0 ) return;

    // Top row
    int y, row, voff = vscrollbar->value();
    for ( row=y=0; row < _rows; row++ )
    {
        y += row_height(row);
        if ( y >= voff ) { y -= row_height(row); break; }
    }
    toprow = ( row >= _rows ) ? (row - 1) : row;

    // Bottom row
    voff = vscrollbar->value() + tablebox->h();
    for ( ; row < _rows; row++ )
    {
        y += row_height(row);
        if ( y >= voff ) { break; }
    }
    botrow = ( row >= _rows ) ? (row - 1) : row;

    // Left column
    int x, col, hoff = hscrollbar->value();
    for ( col=x=0; col < _cols; col++ )
    {
        x += col_width(col);
        if ( x >= hoff ) { x -= col_width(col); break; }
    }
    leftcol = ( col >= _cols ) ? (col - 1) : col;

    // Right column
    //    Work with data left over from leftcol calculation
    //
    hoff = hscrollbar->value() + tablebox->w();
    for ( ; col < _cols; col++ )
    {
        x += col_width(col);
        if ( x >= hoff ) { break; }
    }
    rightcol = ( col >= _cols ) ? (col - 1) : col;
}

// Table resized: recalc internal data
//    Call this whenever the window is resized.
//    Recalculates the scrollbar sizes.
//    Makes no assumptions about any pre-initialized data.
//
void Fl_Table::table_resized()
{
    _tableh = row_scroll_position(rows());
    _tablew = col_scroll_position(cols());

    //
    // Make scroll bars disappear if window large enough
    //

    // First pass: can hide via window size?
    int hidev = ( _tableh - h() < 0 ),
	hideh = ( _tablew - w() < 0 );

    // Second pass: Check for interference
    if ( !hideh & hidev ) { hidev = ( ( _tableh - h() + SCROLLBAR_SIZE ) < 0 ); } 
    if ( !hidev & hideh ) { hideh = ( ( _tablew - w() + SCROLLBAR_SIZE ) < 0 ); }

    // Determine scrollbar visibility
    if ( hidev ) { vscrollbar->hide(); } else { vscrollbar->show(); }
    if ( hideh ) { hscrollbar->hide(); } else { hscrollbar->show(); }

    //
    // Get the table's physical clipping window.
    //
    //    Do this AFTER hide/showing scrollbars
    //    since if they're shown, they affect 
    //    the size of the table window.
    //
    int twx, twy, tww, twh;
    table_window_size(twx, twy, tww, twh);
    tablebox->resize(twx, twy, tww, twh);
    tablebox->redraw();

    // Recalc scrollbar sizes
    //    Clamp scrollbar value() after a resize.
    //    Resize scrollbars to enforce a constant trough width after a window resize.
    //
    {
        float vscrolltab = ( _tableh == 0 || twh > _tableh ) ? 1 : (float)twh / _tableh;
        float hscrolltab = ( _tablew == 0 || tww > _tablew ) ? 1 : (float)tww / _tablew;

	vscrollbar->bounds(0, _tableh-twh);
	vscrollbar->precision(10);
	vscrollbar->slider_size(vscrolltab);
	vscrollbar->resize(x()+w()-SCROLLBAR_SIZE, y(), 
			   SCROLLBAR_SIZE, h() - ((hideh)?0:SCROLLBAR_SIZE));
	vscrollbar->Fl_Valuator::value(vscrollbar->clamp(vscrollbar->value()));	

	hscrollbar->bounds(0, _tablew-tww);
	hscrollbar->precision(10);
        hscrollbar->slider_size(hscrolltab);
	hscrollbar->resize(x(), y()+h()-SCROLLBAR_SIZE,
			   w() - ((hidev)?0:SCROLLBAR_SIZE), SCROLLBAR_SIZE);
	hscrollbar->Fl_Valuator::value(hscrollbar->clamp(hscrollbar->value()));
    }

    // Tell FLTK child widgets were resized
    Fl_Group::init_sizes();

    // Recalc top/bot/left/right
    table_scrolled();

    // Redraw everything
    redraw();
}

// Someone moved a scrollbar
void Fl_Table::scroll_cb(Fl_Widget*w, void *data)
{
    Fl_Table *o = (Fl_Table*)data;
    o->table_scrolled();
    o->redraw();
}

// Set number of rows
void Fl_Table::rows(int val)
{
    _rows = val;
    {
	int default_h = ( _rowheights.size() > 0 ) ? _rowheights.back() : 25;
	while ( val > (int)_rowheights.size() ) { _rowheights.push_back(default_h); }	// enlarge
	while ( val < (int)_rowheights.size() ) { _rowheights.pop_back(); }		// shrink
    }
    table_resized();
}

// Set number of cols
void Fl_Table::cols(int val)
{
    _cols = val;
    {
	int default_w = ( _colwidths.size() > 0 ) ? _colwidths.back() : 80;
	while ( val > (int)_colwidths.size() ) { _colwidths.push_back(default_w); }	// enlarge
	while ( val < (int)_colwidths.size() ) { _colwidths.pop_back(); }		// shrink
    }
    table_resized();
}

// Change mouse cursor to different type
void Fl_Table::change_cursor(Fl_Cursor newcursor)
{
    if ( newcursor != last_cursor )
    {
	fl_cursor(newcursor, FL_BLACK, FL_WHITE);
	last_cursor = newcursor;
    }
}

// #include "eventnames.h"

// Handle FLTK events
int Fl_Table::handle(int event)
{
    // Insure scrollbars are last children, for correct event handling
    fix_scrollbar_order();

    // fprintf(stderr, "** EVENT: %s: ", eventnames[event]);

    static int resizing_col = -1,			// column being dragged
	       resizing_row = -1,			// row being dragged
	       dragging_x = -1,				// starting x position for horiz drag
	       dragging_y = -1,				// starting y position for vert drag
	       push_row = -1,				// FL_PUSH on this row (used by FL_RELEASE)
	       push_col = -1;				// FL_PUSH on this col (used by FL_RELEASE)

    static TableContext push_context = CONTEXT_NONE;	// FL_PUSH context

    int ret = Fl_Group::handle(event);	// let FLTK group handle events first

    // Which row/column are we over?
    int R, C;  				// row/column being worked on
    ResizeFlag resizeflag;		// which resizing area are we over? (0=none)
    TableContext context = cursor2rowcol(R, C, resizeflag);

    switch ( event )
    {
	case FL_PUSH:

	    // Save these for later when button released.
	    //    Do this for all mouse push buttons.
	    //
	    push_context = context;
	    push_row = R;
	    push_col = C;

	    if ( context == CONTEXT_COL_HEADER && Fl::event_button() == 1 && resizeflag )
	    {
		// Start resize if left mouse clicked on column border.
		//
		//    "ret = 1" ensures we get drag events from now on.
		//    (C-1) is used if mouse is over the left hand side of the cell,
		//    so that we resize the next column over to the left.
		//
		resizing_col = ( resizeflag & RESIZE_COL_LEFT ) ? C-1 : C; 
		dragging_x = Fl::event_x(); 
		ret = 1;
	    }

	    if ( context == CONTEXT_ROW_HEADER && Fl::event_button() == 1 && resizeflag )
	    {
		// Start resize if left mouse clicked on row border.
		//
		//    "ret = 1" ensures we get drag events from now on.
		//    (R-1) is used if mouse is over the top of the cell,
		//    so that we resize the row above.
		//
		resizing_row = ( resizeflag & RESIZE_ROW_ABOVE ) ? R-1 : R; 
		dragging_y = Fl::event_y(); 
		ret = 1;
	    }

	    break;

        case FL_DRAG:

	    // Dragging column?
	    //
	    //    Let user drag even /outside/ the row/col widget.
	    //    Don't allow column width smaller than 1.
	    //    Continue to show FL_CURSOR_WE at all times during drag.
	    //
	    if ( resizing_col > -1 )
	    {
		int offset = dragging_x - Fl::event_x();
		int new_w = col_width(resizing_col) - offset;
		if ( new_w < 1 ) new_w = 1;
		col_width(resizing_col, new_w);
		dragging_x = Fl::event_x();
		table_resized();
		change_cursor(FL_CURSOR_WE);
		ret = 1;
	    }

	    // Dragging row?
	    //
	    //    Let user drag even /outside/ the row/col widget.
	    //    Don't allow row width smaller than 1.
	    //    Continue to show FL_CURSOR_NS at all times during drag.
	    //
	    if ( resizing_row > -1 )
	    {
		int offset = dragging_y - Fl::event_y();
		int new_h = row_height(resizing_row) - offset;
		if ( new_h < 1 ) new_h = 1;
		row_height(resizing_row, new_h);
		dragging_y = Fl::event_y();
		table_resized();
		change_cursor(FL_CURSOR_NS);
		ret = 1;
	    }
	    break;

	case FL_RELEASE:
	    switch ( context )
	    {
		case CONTEXT_COL_HEADER:	// release on header
		case CONTEXT_CELL:		// release on a cell

		    // FL_RELEASE on same row/col we FL_PUSH'ed?
		    //    Then do the user's callback with context.
		    //
		    if ( _when & (FL_WHEN_RELEASE | FL_WHEN_RELEASE_ALWAYS ) &&	// when qualifies?
		         push_context == context && 			// same context?
		         push_row == R && 				// same row FL_PUSH'ed?
			 push_col == C && 				// same col FL_PUSH'ed?
			 resizing_col == -1 )				// not resizing a column
		    {
		        callback(context, R, C);	// don't use Fl_Table::callback(),
							// or parent class can't override
		    }
		    break;

		default:
		    break;
	    }

	    if ( Fl::event_button() == 1 )
	    {
		change_cursor(FL_CURSOR_DEFAULT);
		resizing_col = -1;
		resizing_row = -1;
		ret = 1;
	    }
	    break;

	case FL_MOVE:
	    if ( context == CONTEXT_COL_HEADER && 	// in column header?
	         resizeflag )				// and resize enabled + near resize boundary?
		{ change_cursor(FL_CURSOR_WE); }	// show resize cursor
	    else if ( context == CONTEXT_ROW_HEADER && 	// in row header?
	         resizeflag )				// and resize enabled + near resize boundary?
		{ change_cursor(FL_CURSOR_NS); }	// show resize cursor
	    else
		{ change_cursor(FL_CURSOR_DEFAULT); }	// normal cursor
	    ret = 1;
	    break;

	default:
	    change_cursor(FL_CURSOR_DEFAULT);
	    break;
    }

    return(ret);
}

// Resize FLTK override
//     Handle resize events if user resizes parent window.
//     I YAM HERE: If X or Y is changed, must move children's x/y's too! See Fl_Scroll::resize()
//
void Fl_Table::resize(int X, int Y, int W, int H)
{
    // Tell group to resize, and recalc our own widget as well
    Fl_Group::resize(X, Y, W, H);
    table_resized();
}

// Draw the entire Fl_Table
//    Override the draw() routine to draw the table.
//    Then tell the group to draw over us.
//
void Fl_Table::draw()
{
    int cx, cy, cw, ch;	// clip box
    get_bounds(CONTEXT_TABLE, cx, cy, cw, ch);

    draw_cell(CONTEXT_NEWPAGE);		// let user's drawing routine prepare for new page

    // Draw widget's box
    //    The widget border /inside/ the scroll bars, if any
    //
    if ( damage() & FL_DAMAGE_ALL || damage() & FL_DAMAGE_CHILD )
	{ draw_box(box(), cx, cy, cw, ch, color()); }

    int X,Y,W,H;

    // Draw row headers, if any
    if ( row_header() )
    {
	// Draw row header left of 'tablebox'; compensate x clip
	fl_push_clip(tablebox->x()-row_header_width(), 
		     tablebox->y(),
		     row_header_width(),
		     tablebox->h());
	for ( int r = toprow; r <= botrow; r++ )
	{
	    find_cell(CONTEXT_ROW_HEADER, r, 0, X, Y, W, H);
	    draw_cell(CONTEXT_ROW_HEADER, r, 0, X, Y, W, H);
	}
	fl_pop_clip();
    }

    // Draw column headers, if any
    if ( col_header() )
    {
	// Draw column header *above* 'tablebox'; compensate y clip
	fl_push_clip(tablebox->x(), 
		     tablebox->y()-col_header_height(), 
		     tablebox->w(), 
		     col_header_height());

	for ( int c = leftcol; c <= rightcol; c++ )
	{
	    find_cell(CONTEXT_COL_HEADER, 0, c, X, Y, W, H);
	    draw_cell(CONTEXT_COL_HEADER, 0, c, X, Y, W, H);
	}
	fl_pop_clip();
    }

    // Draw all cells.
    //    This includes cells partially obscured off edges of table.
    //
    fl_push_clip(cx, cy, cw, ch);
    {
	for ( int r = toprow; r <= botrow; r++ )
	{
	    for ( int c = leftcol; c <= rightcol; c++ )
	    {
		find_cell(CONTEXT_CELL, r, c, X, Y, W, H);
		draw_cell(CONTEXT_CELL, r, c, X, Y, W, H);
	    }
	}
    }
    fl_pop_clip();

    // Draw little rectangle in corner of headers
    if ( row_header() && col_header() )
    {
        fl_color(color());
	fl_rectf(x(), y(), row_header_width(), col_header_height());
    }

    // Window wider than table? Fill remainder with rectangle
    if ( cw > _tablew )
    {
        fl_color(color());
        fl_rectf(tablebox->x() + _tablew, tablebox->y(),
	         tablebox->w() - _tablew, tablebox->h());

	// Col header? fill that too
	fl_color(col_header_color());
	fl_rectf(tablebox->x() + _tablew, tablebox->y() - col_header_height(),
		 tablebox->w() - _tablew, col_header_height());
    }

    // Window higher than table? Fill remainder with rectangle
    if ( ch > _tableh )
    {
        fl_color(color());
	fl_rectf(tablebox->x(), tablebox->y() + _tableh,
	         tablebox->w(), tablebox->h() - _tableh);

	// Row header? fill that too
	fl_color(row_header_color());
	fl_rectf(tablebox->x() - row_header_width(), tablebox->y() + _tableh,
		 row_header_width(),                 tablebox->h() - _tableh);
    }

    // Draw scrollbars separately
    if ( damage() & FL_DAMAGE_ALL )
    {
	if ( vscrollbar->visible() ) 
	    { Fl_Widget *w = (Fl_Widget*)vscrollbar; draw_child(*w); }
	if ( hscrollbar->visible() )
	    { Fl_Widget *w = (Fl_Widget*)hscrollbar; draw_child(*w); }

	// fill in little box in corner
	if ( vscrollbar->visible() && hscrollbar->visible() )
	{
	    fl_color(color());
	    fl_rectf(vscrollbar->x(), hscrollbar->y(), 
	             vscrollbar->w(), hscrollbar->h());
	}
    }
    else
    {
        update_child(*vscrollbar);
        update_child(*hscrollbar);
    }

    // Draw user's child fltk widgets, if we're a container
    if ( is_fltk_container() )
    {
	fl_push_clip(cx+1, cy+1, cw-2, ch-2);
	Fl_Group::draw();
	fl_pop_clip();
    }
}

// Print internal variables for debugging
void Fl_Table::printself()
{
    cerr << "------------------------------------------------" << endl
	 << "        _rows/_cols: " << _rows << "," << _cols << endl
	 << "      _row_header_w: " << _row_header_w << endl
	 << "      _col_header_h: " << _col_header_h << endl
	 << "        _row_header: " << (int)_row_header << endl
	 << "        _col_header: " << (int)_col_header << endl
	 << "        _row_resize: " << (int)_row_resize << endl
	 << "        _col_resize: " << (int)_col_resize << endl
	 << "  _colwidths.size(): " << _colwidths.size() << endl
	 << " _rowheights.size(): " << _rowheights.size() << endl
	 << "        last_cursor: " << (int)last_cursor << endl
	 << "              _when: " << (int)_when << endl
	 << "            _tablew: " << _tablew << endl
	 << "            _tableh: " << _tableh << endl
	 << "             toprow: " << toprow << endl
	 << "             botrow: " << botrow << endl
	 << "            leftcol: " << leftcol << endl
	 << "           rightcol: " << rightcol << endl
	 << "           tablebox: " << tablebox->x() << "," << tablebox->y() << ","
				    << tablebox->w() << "," << tablebox->h() << endl
	 << "         vscrollbar: " << vscrollbar->x() << "," << vscrollbar->y() << ","
				    << vscrollbar->w() << "," << vscrollbar->h() << endl
	 << "         hscrollbar: " << hscrollbar->x() << "," << hscrollbar->y() << ","
				    << hscrollbar->w() << "," << hscrollbar->h() << endl
	 << "------------------------------------------------" << endl
	 << endl;
}
